"""
Module to implement a generic representation of media objects
that can be posted or retrieved as part of messages
"""

from dataclasses import dataclass
from typing import Final

SUPPORTED_MIME_TYPES: Final[set[str]] = {
    "image/jpeg",
    "image/png",
    "image/gif",
    "image/webp",
    "video/mp4",
    "video/quicktime",
}


@dataclass(frozen=True)
class Media:
    """
    A generic media object with a MIME type and the content
    in bytes.

    Can be retrieved as part of a message or posted
    to a social media network.
    """

    # Required attributes to represent a media object
    mime_type: str
    content: bytes

    # Optional metadata
    alt_text: str = ""

    def is_valid(self) -> bool:
        """
        Check if the media object is valid and can be
        posted to a social media network or otherwise
        handled.

        For a media object to be valid, it must have
        a non-empty content and a valid MIME type.
        """

        has_content = len(self.content) > 0
        has_mime_type = self.mime_type.strip()

        if not has_content or not has_mime_type:
            return False

        return self.mime_type in SUPPORTED_MIME_TYPES



"""
Module to implement a custom connection class for Bluesky accounts,
supporting reading and writing statuses from the authenticated user
via their handle and password.
"""

import io
import logging
import time
from datetime import datetime, timezone
from typing import Final, Optional, Union
from urllib.parse import urlparse

import requests
from atproto import Client
from atproto_client.exceptions import (  # type: ignore
    BadRequestError,
    InvokeTimeoutError,
    RequestException,
)
from atproto_client.models import (  # type: ignore
    AppBskyEmbedExternal,
    AppBskyEmbedImages,
    AppBskyEmbedRecord,
    AppBskyEmbedRecordWithMedia,
    AppBskyEmbedVideo,
    AppBskyRichtextFacet,
)
from atproto_client.models.blob_ref import BlobRef  # type: ignore
from atproto_client.models.common import XrpcError  # type: ignore
from atproto_client.models.string_formats import Did  # type: ignore
from atproto_client.namespaces.sync_ns import ComAtprotoSyncNamespace  # type: ignore
from bs4 import BeautifulSoup, Tag
from PIL import Image

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Media, Message, MessageType
from barkr.utils import (
    REQUESTS_EMBED_GET_TIMEOUT,
    REQUESTS_HEADERS,
    extract_urls_from_text,
)

logger = logging.getLogger()

BLUESKY_MAX_MESSAGE_LENGTH: Final[int] = 300
BLUESKY_MAX_IMAGE_SIZE_BYTES: Final[int] = 1000000


class BlueskyConnection(Connection):
    """
    Custom connection class for Bluesky accounts,
    supporting reading and writing statuses from the authenticated user.

    Requires handle and an app password for authentication.
    """

    supported_message_type: MessageType = MessageType.TEXT_MEDIA
    service: Client
    handle: str
    compress_images: bool

    def __init__(  # pylint: disable=too-many-arguments too-many-positional-arguments
        self,
        name: str,
        modes: list[ConnectionMode],
        handle: str,
        password: str,
        compress_images: bool = False,
    ) -> None:
        """
        Initializes the connection with a name and a list of modes
        and sets up the initial connection between the client and Bluesky
        for the given user.

        NOTE: it is recommended to use an app password instead of the user's password.
        ref: https://bsky.app/settings/app-passwords

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param handle: The handle of the authenticated user
        :param password: The app password of the authenticated user
        :param compress_images: Whether to compress images before posting
        """

        super().__init__(name, modes)

        logger.info(
            "Initializing Bluesky (%s) connection for user %s",
            self.name,
            handle,
        )

        self.service = Client()
        self.service.login(handle, password)
        self.handle = handle
        self.compress_images = compress_images

        logger.info(
            "Bluesky (%s) connection initialized! (User handle: %s)",
            self.name,
            self.handle,
        )

        user_feed = self.service.app.bsky.feed.get_author_feed({"actor": handle}).feed
        if user_feed:
            # Set the initial min_id to the most recent post's indexed_at,
            # which is a UTC timestamp string, casted to datetime
            self.min_id: Optional[datetime] = datetime.fromisoformat(
                user_feed[0].post.indexed_at
            )
            logger.info("Bluesky (%s) initial min_id: %s", self.name, self.min_id)
        else:
            self.min_id = None
            logger.info("Bluesky (%s) initial min_id not set.", self.name)

    def _fetch(self) -> list[Message]:
        """
        Fetches messages from the authenticated user's account.

        :return: A list of messages
        """

        messages: list[Message] = []

        user_feed = self.service.app.bsky.feed.get_author_feed(
            {"actor": self.handle}
        ).feed
        if user_feed:
            for feed_view in user_feed:
                post = feed_view.post

                # Ignoring reposts
                if post.viewer is not None and post.viewer.repost is not None:
                    continue

                # Ignoring replies
                if post.record.reply is not None:
                    continue

                if (
                    self.min_id is None
                    or datetime.fromisoformat(post.indexed_at) > self.min_id
                ):
                    record = post.record
                    if (embed := record.embed) is not None:
                        text = self._process_text_with_embed(record.text, embed)
                    else:
                        text = record.text

                    language = None
                    if record.langs:
                        language = record.langs[0]

                    media_list = self._extract_media_list_from_embed(
                        post.author.did, embed
                    )
                    messages.append(
                        Message(
                            id=post.indexed_at,
                            message=text,
                            language=language,
                            media=media_list,
                        )
                    )

        if messages:
            self.min_id = datetime.fromisoformat(messages[0].id)
            logger.info("Bluesky (%s) has %s new messages.", self.name, len(messages))
        else:
            logger.info("Bluesky (%s) has no new messages.", self.name)

        return messages

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Posts the given messages to the authenticated user's account.

        :param messages: The messages to post
        :return: A list of IDs of the posted messages
        """

        posted_message_ids: list[str] = []

        for message in messages:
            if len(message.message) > BLUESKY_MAX_MESSAGE_LENGTH:
                logger.warning(
                    "Message length exceeds Bluesky (%s) "
                    "maximum length (%s), skipping: %s",
                    self.name,
                    BLUESKY_MAX_MESSAGE_LENGTH,
                    message.message,
                )
                continue

            embed, facets = self._generate_post_embed_and_facets(message.message)
            language = [message.language] if message.language else None
            try:
                created_record = self.service.send_post(
                    text=message.message,
                    embed=embed,
                    facets=facets if facets else None,
                    langs=language,
                )
            except InvokeTimeoutError as e:
                # Something happened with the Bluesky API, let's recover
                logger.error(
                    "Bluesky (%s) post failed with timeout error: %s", self.name, e
                )

                # In case we _did_ post the message, we don't want to
                # re-post it again
                self.min_id = _get_current_indexed_at()
                continue
            except BadRequestError as e:
                # We could be trying to create an embed that is too large,
                # let's recover
                error_response = e.response
                content = error_response.content

                if isinstance(content, XrpcError):
                    self._retry_post_message(message, facets, language, e)
                else:
                    logger.error(
                        "Bluesky (%s) post failed with unexpected error: %s",
                        self.name,
                        content,
                    )
                    raise e

            created_uri = created_record.uri

            # NOTE: introducing an artificial delay to ensure the post is indexed
            # before fetching the post details
            time.sleep(2)

            try:
                post_details = self.service.get_posts([created_uri]).posts[0]
                indexed_at = datetime.fromisoformat(post_details.indexed_at)
            except IndexError:
                indexed_at = _get_current_indexed_at()
                logger.warning(
                    "Failed to fetch post details for Bluesky (%s) post: %s, "
                    "manually setting indexed_at to %s",
                    self.name,
                    created_uri,
                    indexed_at,
                )

            logger.info(
                "Posted message %s to Bluesky (%s) connection (URI: %s, Indexed At: %s)",
                message.message,
                self.name,
                created_uri,
                indexed_at,
            )

            self.min_id = indexed_at
            posted_message_ids.append(str(indexed_at))

        return posted_message_ids

    def _retry_post_message(
        self,
        message: Message,
        facets: Optional[list[AppBskyRichtextFacet.Main]],
        language: Optional[list[str]],
        bad_request_error: BadRequestError,
    ) -> None:
        """
        Given a Bluesky XRPC error, attempts to retry posting the message
        if the error is recoverable. In most cases, this will be due to
        an error with the embed (size, mime type...) so we retry
        without it.

        :param message: The message to retry posting
        :param facets: The facets to attach to the post
        :param language: The language of the post
        :param xrcp_error: The XRPC error that occurred
        """

        xrcp_error = bad_request_error.response.content
        if not isinstance(xrcp_error, XrpcError):
            logger.error(
                "Unexpected non-XRPC error when posting to Bluesky (%s): %s",
                self.name,
                bad_request_error.response.content,
            )
            raise bad_request_error

        if xrcp_error.error == "BlobTooLarge":
            logger.warning(
                "Bluesky (%s) post failed due to embed size, "
                "reattempting post without embed.",
                self.name,
            )
        elif xrcp_error.error == "InvalidMimeType":
            logger.warning(
                "Bluesky (%s) post failed due to invalid MIME type (%s) "
                "reattempting post without embed.",
                self.name,
                xrcp_error.message,
            )
        else:
            logger.error(
                "Bluesky (%s) post failed with unexpected error: %s",
                self.name,
                xrcp_error,
            )
            raise bad_request_error

        self.service.send_post(
            text=message.message,
            embed=None,
            facets=facets if facets else None,
            langs=language,
        )

    def _generate_post_embed_and_facets(
        self, text: str
    ) -> tuple[Optional[AppBskyEmbedExternal.Main], list[AppBskyRichtextFacet.Main]]:
        """
        If a link is present on the text for a post to be created,
        creates an Embed object for the link.

        This allows Bluesky to display the link as a preview.

        Also generate facets for the link to be used in the post.

        :param text: The text of the post
        :return: A tuple containing the Embed object and a list of facets
        """

        embed = None
        facets: list[AppBskyRichtextFacet.Main] = []

        for url in extract_urls_from_text(text):
            # Create a facet for the link. We do this first to always
            # have a facet for the link, even if we fail to create the embed
            # (e.g. if we can't fetch the URL metadata)
            url_index = text.index(url)
            facets.append(
                AppBskyRichtextFacet.Main(
                    features=[
                        AppBskyRichtextFacet.Link(
                            uri=url,
                        )
                    ],
                    index=AppBskyRichtextFacet.ByteSlice(
                        byte_start=len(text[:url_index].encode("utf-8")),
                        byte_end=len(text[: url_index + len(url)].encode("utf-8")),
                    ),
                )
            )

            # If we already have an embed, we don't need to create another one
            # for the same post
            if embed is not None:
                continue

            # If we haven't created an embed yet, hit the URL to check if it
            # is valid and extract title, thumbnail and description
            try:
                # Using a very short timeout, we don't want to spend too much time here
                response = requests.get(
                    url, timeout=REQUESTS_EMBED_GET_TIMEOUT, headers=REQUESTS_HEADERS
                )
            except requests.RequestException:
                continue

            if response.status_code == 200:
                # Parse the response to extract title, description and image
                soup = BeautifulSoup(response.content, "html.parser")
                title = soup.title.string if soup.title else urlparse(url).netloc
                description = url
                thumbnail_blob = None

                # Extract the description from the meta tag
                if (
                    meta_description := _get_meta_tag_from_html_metadata(
                        soup, "og:description"
                    )
                ) is not None:
                    description = meta_description

                # Extract the image from the meta tag
                if (
                    image := _get_meta_tag_from_html_metadata(soup, "og:image")
                ) is not None:
                    # Fetching the image and reuploading to Bluesky
                    thumbnail_blob = self._upload_image_url_to_atproto_blob(image)

                # Prepare the Embed object
                embed = AppBskyEmbedExternal.Main(
                    external=AppBskyEmbedExternal.External(
                        uri=url,
                        title=title,
                        description=description,
                        thumb=thumbnail_blob,
                    )
                )

        return embed, facets

    def _extract_media_list_from_embed(
        self,
        did: Did,
        embed: Optional[
            Union[
                AppBskyEmbedExternal.Main,
                AppBskyEmbedRecord.Main,
                AppBskyEmbedImages.Main,
                AppBskyEmbedVideo.Main,
                AppBskyEmbedRecordWithMedia.Main,
            ]
        ],
    ) -> list[Media]:
        """
        Given a record's embed object, extracts the media list from
        any attached images or videos.

        :param embed: The embed object containing the media
        :return: A list of Media objects
        """

        media_list: list[Media] = []

        # If there is no embed, we return an empty list
        if embed is None:
            return media_list

        blob_refs_with_alt_text: list[tuple[BlobRef, str]] = []

        # Extracting the blob references from the embed
        # per embed type
        if isinstance(embed, AppBskyEmbedVideo.Main):
            blob_refs_with_alt_text.append((embed.video, embed.alt or ""))

        elif isinstance(embed, AppBskyEmbedImages.Main):
            blob_refs_with_alt_text.extend(
                [(image.image, image.alt) for image in embed.images]
            )

        # Iterate over the blob references and fetch the blobs with their
        # respective MIME types; note that if the embed was of a different
        # type, this iteration should be skipped
        for blob_ref, alt_text in blob_refs_with_alt_text:
            if blob_ref is None:
                continue

            mime_type = blob_ref.mime_type
            if mime_type is None:
                continue

            blob_cid = blob_ref.cid.encode()
            logger.info(
                "Fetching blob from Bluesky (%s) for CID %s",
                self.name,
                blob_cid,
            )

            # Fetching the blob from Bluesky
            try:
                blob_bytes: bytes = ComAtprotoSyncNamespace(self.service).get_blob(
                    params={"cid": blob_cid, "did": did}
                )
            except (InvokeTimeoutError, BadRequestError, RequestException) as e:
                logger.warning(
                    "Failed to fetch blob from Bluesky (%s) for CID %s: %s",
                    self.name,
                    blob_cid,
                    e,
                )
                continue

            if blob_bytes is not None:
                media_list.append(
                    Media(
                        mime_type=mime_type,
                        content=blob_bytes,
                        alt_text=alt_text,
                    )
                )

        return media_list

    def _process_text_with_embed(
        self,
        text: str,
        embed: Optional[
            Union[
                AppBskyEmbedExternal.Main,
                AppBskyEmbedRecord.Main,
                AppBskyEmbedImages.Main,
                AppBskyEmbedVideo.Main,
                AppBskyEmbedRecordWithMedia.Main,
            ]
        ],
    ) -> str:
        """
        Handles the special case where a Bluesky post contains a link to an embedded
        resources that is not fully rendered as part of the text.

        Leveraging the Embed object, reconstructs the text to include
        the full URL to the resource.

        For example, when posting the URL
        https://open.spotify.com/track/0ElVpg9XIswx3XWs6kUj6a?si=0015d86587524ef9
        the text is trimmed to open.spotify.com/track/0ElVpg... but the
        Embed object contains the full URL.

        :param text: The original text of the post
        :param embed: The Embed object containing the link
        :return: The reconstructed text with the full URL
        """

        if embed is None:
            return text

        # Depending on the type of embed, we get the URL
        # from the corresponding field
        if isinstance(embed, AppBskyEmbedExternal.Main):
            url = embed.external.uri
        else:
            return text

        # We now want to find the word in the text that is contained
        # in the URL, and we only care for the _longest_ word
        # if there are multiple matches
        matching_word = ""
        for word in text.split():
            if word.replace("...", "") in url:
                if len(word) > len(matching_word):
                    matching_word = word

        if not matching_word:
            return text

        return text.replace(matching_word, url)

    def _upload_image_url_to_atproto_blob(self, image_url: str) -> Optional[BlobRef]:
        """
        Given a URL to an image, fetches the image and uploads it
        to Bluesky as a blob. If the image is larger than the maximum
        allowed size, it will be compressed.

        :param image_url: The URL of the image to upload
        :return: The BlobRef object referencing the uploaded image blob
        """

        try:
            img_data = requests.get(
                image_url,
                timeout=REQUESTS_EMBED_GET_TIMEOUT,
                headers=REQUESTS_HEADERS,
            ).content
        except requests.RequestException as e:
            logger.warning("Failed to fetch image from %s: %s", image_url, e)
            return None

        # Check if image needs compression
        if len(img_data) > BLUESKY_MAX_IMAGE_SIZE_BYTES:
            logger.info(
                "Image from %s is %d bytes, exceeds limit of %d bytes.",
                image_url,
                len(img_data),
                BLUESKY_MAX_IMAGE_SIZE_BYTES,
            )

            if not self.compress_images:
                logger.warning(
                    "Image compression is disabled for Bluesky (%s), "
                    "skipping upload for %s",
                    self.name,
                    image_url,
                )
                return None

            img_data = self._compress_image(img_data)
            if img_data is None:
                logger.warning("Failed to compress image from %s", image_url)
                return None

        try:
            return self.service.upload_blob(img_data).blob
        except (InvokeTimeoutError, BadRequestError, RequestException) as e:
            logger.warning("Failed to upload image to Bluesky (%s): %s", self.name, e)
            return None

    def _compress_image(self, img_data: bytes) -> Optional[bytes]:
        """
        Compress an image to fit within the Bluesky image size limit.
        This is attempted by scaling down the image.

        This method uses the Pillow library to handle image processing.
        If compression fails, or the target size cannot be achieved,
        None is returned. This will signal that no image should be uploaded.

        :param img_data: The original image data in bytes
        :return: Compressed image data if successful, None otherwise
        """

        try:
            # Open the image to ensure it's valid
            img = Image.open(io.BytesIO(img_data))

            # If the image is already smaller than the limit,
            # no need to compress it further
            if len(img_data) <= BLUESKY_MAX_IMAGE_SIZE_BYTES:
                logger.info(
                    "Image is already within size limit (%d bytes), "
                    "no compression needed.",
                    len(img_data),
                )
                return img_data

            # Resizing the image to fit within the size limit
            original_width, original_height = img.size

            for scale_factor in [0.8, 0.75]:
                new_width = int(original_width * scale_factor)
                new_height = int(original_height * scale_factor)

                resized_img = img.resize(
                    (new_width, new_height), Image.Resampling.LANCZOS
                )

                # Try different qualities with the resized image
                for quality in [85, 70]:
                    output = io.BytesIO()
                    resized_img.save(
                        output, format="JPEG", quality=quality, optimize=True
                    )
                    compressed_data = output.getvalue()

                    if len(compressed_data) <= BLUESKY_MAX_IMAGE_SIZE_BYTES:
                        logger.info(
                            "Compressed image to %d bytes using "
                            "resize=%dx%d and quality=%d",
                            len(compressed_data),
                            new_width,
                            new_height,
                            quality,
                        )
                        return compressed_data

            # If we still can't compress enough, give up
            logger.warning("Could not compress image to fit within size limit")
            return None

        except Exception as e:  # pylint: disable=broad-except
            logger.warning("Error compressing image: %s", e)
            return None


def _get_meta_tag_from_html_metadata(
    soup: BeautifulSoup, tag_name: str
) -> Optional[str]:
    """
    Extracts the content of meta tag from the HTML metadata of a page.

    If there are multiple meta tags with the same property,
    only the first one is returned.

    :param soup: The BeautifulSoup object containing the HTML metadata
    :param tag_name: The name of the meta tag to extract
    :return: The meta tag content if found, otherwise None
    """

    tag = soup.find("meta", attrs={"property": tag_name})
    if isinstance(tag, Tag) and tag.has_attr("content"):
        tag_content = tag["content"]

        if isinstance(tag_content, list):
            tag_content = tag_content[0]

        return tag_content

    return None


def _get_current_indexed_at() -> datetime:
    """
    Returns the current UTC timestamp,
    to mock the Bluesky API's indexed_at field
    right after posting a message if the post details cannot be fetched.

    :return: The current indexed_at timestamp
    """

    return datetime.now(timezone.utc)



"""
Module to implement a custom connection class for Discord,
supporting writing messages to a Discord channel
via a Discord bot.

This module uses the discord.py library to interact
with the Discord API.
"""

import asyncio
import logging

import discord

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Message

logger = logging.getLogger()


class DiscordConnection(Connection):
    """
    Custom connection class for Discord,
    supporting writing messages to a Discord channel
    via a Discord bot.
    """

    intents: discord.Intents
    token: str
    channel_id: int

    def __init__(
        self, name: str, modes: list[ConnectionMode], token: str, channel_id: int
    ) -> None:
        """
        Initializes the connection with a name and a list of modes, as well
        as the Discord bot token and channel ID.

        The Discord bot must be set up beforehand and be added to the server
        where messages will be posted, with visibility and permissions set
        to view and send messages in the desired channel.

        "Message Content Intent" must be enabled in the Discord Developer Portal
        for the bot to be able to read message content.

        Since Discord.py is usually a blocking library and `barkr` is not
        designed to have connections blocking the main thread, we don't
        create a client on initialization. Instead, we create the client
        as needed to post messages. This might add some overhead,
        but it allows us to keep the main thread free from blocking.

        NOTE: only the write mode is supported. Attempting to use read
        mode will raise a NotImplementedError.

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param token: The token for the Discord bot
        """
        super().__init__(name, modes)

        logger.info("Initializing Discord (%s) connection", self.name)
        if self.modes != [ConnectionMode.WRITE]:
            raise NotImplementedError("DiscordConnection only supports write mode.")

        self.intents = discord.Intents.default()
        self.intents.message_content = True

        self.token = token
        self.channel_id = channel_id

        logger.info("Discord (%s) connection initialized successfully", self.name)

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Post a list of messages to a Discord channel as the authenticated bot.

        :param messages: A list of messages to post
        """

        logger.info(
            "Starting new event loop to send messages to Discord (%s)", self.name
        )
        loop = asyncio.new_event_loop()
        loop.run_until_complete(self._send_messages(messages))
        logger.info("Finished posting messages to Discord (%s)", self.name)

        return []

    async def _send_messages(self, messages: list[Message]) -> None:
        """
        Send a list of messages to a Discord channel as the authenticated bot.

        Creates a new Discord client, authenticates the bot with the provided token,
        and sends messages to the specified channel without blocking the main thread.

        :param messages: A list of messages to send
        """

        client = discord.Client(intents=self.intents)

        @client.event
        async def on_ready():
            logger.info(
                "Discord client connected successfully to send %s messages",
                len(messages),
            )
            channel = client.get_channel(self.channel_id)
            for message in messages:
                await channel.send(message.message)
                logger.info("Message posted to Discord (%s) channel", self.name)

            await client.close()

        await client.start(self.token)



"""
Module to implement generic enums and models to represent
message types.
"""

from enum import Enum


class MessageType(Enum):
    """
    Enum to represent the type of messages a connection supports
    """

    TEXT_ONLY = 1
    TEXT_MEDIA = 2



"""
Module to implement the base class and enums for custom
connections to social media networks, chat services, etc.
"""

import logging
from enum import Enum

from barkr.models import Message, MessageType

logger = logging.getLogger()


class ConnectionMode(Enum):
    """
    Enum to represent connection modes
    """

    READ = 1
    WRITE = 2


class Connection:
    """
    Base class to represent a connection to a social media network,
    chat service, etc.

    Subclasses should implement the _fetch and _post methods, which will
    be called depending on the modes each connection is set up with.

    The base clase already supports a validation to avoid duplicate posting
    of messages already posted by one connection on subsequent fetches.
    """

    name: str
    modes: list[ConnectionMode]
    posted_message_ids: set[str]
    supported_message_type: MessageType = MessageType.TEXT_ONLY

    def __init__(self, name: str, modes: list[ConnectionMode]) -> None:
        """
        Initializes the connection with a name and a list of modes

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        """

        if not modes:
            raise ValueError("At least one mode must be provided for the connection.")

        if len(set(modes)) != len(modes):
            raise ValueError("Duplicate modes are not allowed for the connection.")

        self.name = name
        self.modes = modes
        self.posted_message_ids = set()

    def read(self) -> list[Message]:
        """
        Read and return new messages from this connection

        :return: A list of new messages
        """

        if ConnectionMode.READ not in self.modes:
            return []

        try:
            messages: list[Message] = self._fetch()
        except Exception as e:  # pylint: disable=broad-except
            if isinstance(e, NotImplementedError):
                raise

            logger.error(
                "Error fetching messages from connection %s: %s (%s)",
                self.name,
                e,
                type(e).__name__,
            )
            return []

        new_messages: list[Message] = []
        for message in messages:
            if (status_id := message.id) not in self.posted_message_ids:
                new_messages.append(message)
            else:
                logger.info("Status %s already posted, skipping.", status_id)

        return new_messages

    def write(self, messages: list[Message]) -> None:
        """
        Write messages to this connection

        :param messages: A list of messages to be posted
        """

        if ConnectionMode.WRITE not in self.modes:
            return

        # Discarding any empty messages
        messages_with_content = [
            message
            for message in messages
            if message.has_content(self.supported_message_type)
        ]

        if not messages_with_content:
            logger.warning(
                "All messages are empty (no content) for connection %s", self.name
            )
            return

        if len(messages_with_content) != len(messages):
            logger.warning(
                "Discarded %s empty messages for connection %s",
                len(messages) - len(messages_with_content),
                self.name,
            )

        posted_ids: list[str] = self._post(messages_with_content)
        self.posted_message_ids.update(posted_ids)

    def _fetch(self) -> list[Message]:
        """
        Fetch messages from this connection and returns a list of pairs
        containing (id, message)

        :return: A list of pairs containing (id, message)
        """

        raise NotImplementedError(f"Fetch not implemented for connection {self.name}")

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Post messages to this connection and returns a list of message IDs

        :param messages: A list of messages to be posted
        :return: A list of message IDs
        """

        raise NotImplementedError(f"Post not implemented for connection {self.name}")



"""
Re-exporting all classes from the models submodules
for ease of use.

You can refer to each submodule for more information about the classes,
and how to use them.
"""

from barkr.models.message import Media, Message, MessageType

__all__ = [
    "Media",
    "Message",
    "MessageType",
]



"""
Module to implement a custom connection class for RSS feeds,
supporting reading messages from a given feed.
"""

import logging
from time import struct_time
from typing import Callable, Optional

import feedparser

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Message

logger = logging.getLogger()


def default_feed_message_callback(url: str, title: str) -> str:
    """
    Given the URL and title of an RSS feed entry,
    returns a string that represents the message to be sent.

    The content of the message is formatted as follows:
        <title>: <url>

    :param url: The URL of the feed entry`
    :param title: The title of the feed entry
    :return: a string representation of the feed entry as a message
    """

    return f"{title}: {url}"


class RSSConnection(Connection):
    """
    Custom connection class for RSS feeds,
    supporting reading messages from a given feed.
    """

    feed_url: str
    feed_message_callback: Callable[[str, str], str]
    feed_min_date: Optional[struct_time] = None

    def __init__(
        self,
        name: str,
        modes: list[ConnectionMode],
        feed_url: str,
        feed_message_callback: Callable[
            [str, str], str
        ] = default_feed_message_callback,
    ) -> None:
        """
        Initializes the connection with a name and a list of modes, as well as the
        RSS feed URL.

        NOTE: only the read mode is supported. Attempting to use write mode
        will raise a NotImplementedError.

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param feed_url: The URL of the RSS feed
        """
        super().__init__(name, modes)

        logger.info("Initializing RSS (%s) connection", self.name)
        if self.modes != [ConnectionMode.READ]:
            raise NotImplementedError("RSSConnection only supports read mode.")

        # Storing parameters for later use
        self.feed_url = feed_url
        self.feed_message_callback = feed_message_callback

        # Getting most recent post date
        try:
            feed = feedparser.parse(self.feed_url)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("Error on initial RSS feed fetch : %s", e)
        else:
            if feed.entries:
                # Set the initial min_date to the most recent post's published date
                self.feed_min_date = feed.entries[0].published_parsed
                logger.info(
                    "RSS (%s) initial min_date: %s", self.name, self.feed_min_date
                )
            else:
                logger.info(
                    "RSS (%s) initial min_date not set "
                    "(no entries found on initial fetch).",
                    self.name,
                )

        logger.info("RSS (%s) connection initialized successfully", self.name)

    def _fetch(self) -> list[Message]:
        """
        Fetch messages from the RSS feed.

        :return: A list of messages from the feed
        """
        logger.info("Fetching messages from RSS (%s) connection", self.name)

        try:
            feed = feedparser.parse(self.feed_url)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("Error fetching RSS feed: %s", e)
            return []

        logger.info("Fetched RSS feed successfully!")

        messages = []
        for entry in feed.entries:
            if self.feed_min_date and entry.published_parsed <= self.feed_min_date:
                continue

            message = Message(
                id=entry.link,
                message=self.feed_message_callback(entry.link, entry.title),
            )
            messages.append(message)

        # Updated min_date to the most recent post's published date
        if feed.entries:
            self.feed_min_date = feed.entries[0].published_parsed

        logger.info(
            "Fetched %d messages from RSS (%s) connection", len(messages), self.name
        )
        return messages



"""
Module to implement a generic, wrapper model for messages.

A message is a generic object that can be posted to a social media network
OR retrieved from a social media network, with an ID and a message body, as
well as (potentially) other metadata that each connection can decide
whether or not to take into account.
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from barkr.models.media import Media
from barkr.models.message_type import MessageType


class MessageVisibility(Enum):
    """
    Represents the visibility of a message.
    """

    PUBLIC = 0
    UNLISTED = 1
    PRIVATE = 2
    DIRECT = 3

    @staticmethod
    def from_mastodon_visibility(visibility: str) -> "MessageVisibility":
        """
        Convert a Mastodon visibility string to a MessageVisibility enum.

        :param visibility: The visibility string from Mastodon
        :return: The corresponding MessageVisibility enum
        """
        if visibility == "public":
            return MessageVisibility.PUBLIC

        if visibility == "unlisted":
            return MessageVisibility.UNLISTED

        if visibility == "private":
            return MessageVisibility.PRIVATE

        if visibility == "direct":
            return MessageVisibility.DIRECT

        raise ValueError(f"Unknown Mastodon visibility: {visibility}")

    def to_mastodon_visibility(self) -> str:
        """
        Convert a MessageVisibility enum to a Mastodon visibility string.

        :return: The corresponding visibility string for Mastodon
        """

        if self == MessageVisibility.UNLISTED:
            return "unlisted"

        if self == MessageVisibility.PRIVATE:
            return "private"

        if self == MessageVisibility.DIRECT:
            return "direct"

        return "public"


@dataclass(frozen=True)
class Message:
    """
    A generic message object with an ID and a message body,
    as well as (potentially) other metadata that each connection
    can decide whether or not to take into account.
    """

    # Main attributes
    id: str
    message: str

    # Messages can optionally have media attached to them, which are
    # stored in a list per their byte representation and MIME type.
    media: list[Media] = field(default_factory=list)

    # Optional metadata
    language: Optional[str] = None
    label: Optional[str] = None
    visibility: MessageVisibility = MessageVisibility.PUBLIC

    def has_content(self, supported_message_type: MessageType) -> bool:
        """
        Check if the message has content and therefore can be published.

        At this point, the only check is whether the message is empty or not.
        We also check if the message is direct or private, as these messages
        might not be aimed for public posting.

        In the future, as Messages grow in complexity, this method may be
        extended to check for other conditions (e.g. accounting for
        a message with no text but with an image).

        :return: True if the message has content, False otherwise
        """

        # If the message is private or direct, we don't want to post it
        if self.visibility in (MessageVisibility.PRIVATE, MessageVisibility.DIRECT):
            return False

        has_text = bool(self.message.strip())

        # If we support media, we can post a message with no text
        # as long as it has media
        if supported_message_type == MessageType.TEXT_MEDIA:
            valid_media = [m for m in self.media if m.is_valid()]
            has_media = len(valid_media) > 0

            # If we have media, we can post the message
            return has_text or has_media

        # By default, we only support text messages
        return has_text



"""
Module to implement a custom connection class for Mastodon instances,
supporting reading and writing statuses from the authenticated user
via their access token.
"""

import logging
import mimetypes
from typing import Any, Final, Optional

import requests
from bs4 import BeautifulSoup
from mastodon import Mastodon
from mastodon.errors import MastodonNetworkError
from mastodon.return_types import MediaAttachment, Status

from barkr.connections.base import Connection, ConnectionMode
from barkr.models import Media, Message, MessageType
from barkr.models.message import MessageVisibility
from barkr.utils import REQUESTS_EMBED_GET_TIMEOUT, REQUESTS_HEADERS

logger = logging.getLogger()

MASTODON_WRITE_RETRIES: Final[int] = 3


class MastodonConnection(Connection):
    """
    Custom connection class for Mastodon instances,
    supporting reading and writing statuses from the authenticated user.
    """

    supported_message_type = MessageType.TEXT_MEDIA

    service: Mastodon
    min_id: Optional[str]

    def __init__(
        self,
        name: str,
        modes: list[ConnectionMode],
        access_token: str,
        instance_url: str,
    ) -> None:
        """
        Initializes the connection with a name and a list of modes
        as well as the access token and instance URL.

        Validates the access token and instance URL by attempting to
        verify the credentials of the authenticated user and retrieve
        the account ID.

        :param name: The name of the connection
        :param modes: A list of modes for the connection
        :param access_token: The access token for the authenticated user
        :param instance_url: The URL of the Mastodon instance
        """
        super().__init__(name, modes)

        logger.info(
            "Initializing Mastodon (%s) connection to instance %s",
            self.name,
            instance_url,
        )
        self.service = Mastodon(
            access_token=access_token,
            api_base_url=instance_url,
            request_timeout=10,
        )
        self.account_id: str = self.service.account_verify_credentials()["id"]
        logger.info(
            "Mastodon (%s) connection initialized! (Account ID: %s)",
            self.name,
            self.account_id,
        )

        current_statuses = self.service.account_statuses(
            self.account_id, exclude_reblogs=True, exclude_replies=True
        )
        if current_statuses:
            self.min_id = current_statuses[0]["id"]
            logger.info("Mastodon (%s) initial min_id: %s", self.name, self.min_id)
        else:
            self.min_id = None
            logger.info("Mastodon (%s) initial min_id not set.", self.name)

    def _fetch(self) -> list[Message]:
        """
        Fetch messages from this connection

        :return: A list of messages
        """

        statuses: list[Status] = self.service.account_statuses(
            self.account_id,
            exclude_reblogs=True,
            exclude_replies=True,
            min_id=self.min_id,
        )

        if statuses:
            logger.info(
                "Fetched %s new statuses from Mastodon (%s)", len(statuses), self.name
            )
            self.min_id = statuses[0]["id"]
        else:
            logger.info("No new statuses fetched from Mastodon (%s)", self.name)

        return [
            Message(
                id=status["id"],
                message=BeautifulSoup(status["content"], "lxml").text,
                language=status["language"],
                label=status["spoiler_text"] or None,
                visibility=MessageVisibility.from_mastodon_visibility(
                    status["visibility"]
                ),
                media=_get_media_list_from_status(status),
            )
            for status in statuses
            if status["in_reply_to_id"] is None and status["reblog"] is None
        ]

    def _post(self, messages: list[Message]) -> list[str]:
        """
        Post messages from a list to the Mastodon instance

        :param messages: A list of messages to be posted
        :return: A list of message IDs
        """

        posted_message_ids: list[str] = []

        for message in messages:
            attempts = 0

            media_list = _post_media_list_to_mastodon(self.service, message.media)

            while attempts < MASTODON_WRITE_RETRIES:
                try:
                    posted_message = self.service.status_post(
                        message.message,
                        language=message.language,
                        spoiler_text=message.label or "",
                        visibility=message.visibility.to_mastodon_visibility(),
                        media_ids=media_list,
                    )
                except MastodonNetworkError as e:
                    if attempts < MASTODON_WRITE_RETRIES - 1:
                        logger.warning(
                            "Mastodon network error posting status to "
                            "Mastodon (%s), will retry. Error: %s",
                            self.name,
                            e,
                        )
                        attempts += 1
                    else:
                        logger.error(
                            "Failed to post status to Mastodon (%s) "
                            "after %s attempts: %s",
                            self.name,
                            MASTODON_WRITE_RETRIES,
                            e,
                        )
                        raise
                else:
                    break

            posted_message_ids.append(posted_message["id"])
            logger.info(
                "Posted status to Mastodon (%s): %s", self.name, message.message
            )

        return posted_message_ids


def _post_media_list_to_mastodon(
    service: Mastodon, media_list: list[Media]
) -> list[MediaAttachment]:
    """
    Helper function to post a list of Media objects to Mastodon.

    :param media_list: A list of Media objects
    :return: A list of MediaAttachment objects
    """

    media_attachments: list[MediaAttachment] = []

    for media in media_list:
        if not media.is_valid():
            logger.warning("Invalid media object, skipping posting to Mastodon")
            continue

        # Uploading the media to Mastodon
        try:
            media_attachment = service.media_post(
                media_file=media.content,
                mime_type=media.mime_type,
                description=media.alt_text,
            )
        except requests.RequestException as e:
            logger.error("Failed to upload media: %s", e)
            continue

        media_attachments.append(media_attachment)

    return media_attachments


def _get_media_list_from_status(status: dict[str, Any]) -> list[Media]:
    """
    Helper function to extract Media object instances from
    a Mastodon status, if they exist.

    :param status: The Mastodon status object
    :return: A list of Media objects
    """

    media_list: list[Media] = []

    for media_attachment in status["media_attachments"]:
        media_type = media_attachment["type"]

        if media_type not in ["image", "video"]:
            logger.warning(
                "Unsupported media type %s in status %s",
                media_type,
                status["id"],
            )
            continue

        url = media_attachment["url"]

        # Downloading the content of the media from its URL
        try:
            response = requests.get(
                url, timeout=REQUESTS_EMBED_GET_TIMEOUT, headers=REQUESTS_HEADERS
            )
        except requests.RequestException as e:
            logger.error("Failed to download media from %s: %s", url, e)
            continue

        if response.status_code == 200:
            media_content = response.content

            # Guessing the MIME type from the URL
            mime_type, _ = mimetypes.guess_type(url)

            if mime_type is None:
                logger.warning(
                    "Could not determine MIME type for media from URL %s", url
                )
                continue

            alt_text = media_attachment["description"] or ""
            media = Media(
                mime_type=mime_type,
                content=media_content,
                alt_text=alt_text,
            )

            if media.is_valid():
                media_list.append(media)

    return media_list
